#include <fork.h>
#include <io.h>
#include <assert.h>
/* 
See: Practical PRAM Programming with Fork95, A Tutorial
by C. Kessler. Nr. 97-2. pags 16 and 17.
"Self-Scheduling for dependent loop iterations"
implementation by casiano@ull.es
*/
#define MAXSIZE 4096 
typedef int vector[MAXSIZE] ;

sh vector A,B,WAIT;
sh vector D;
sh int N, iteration, slowest, distance, arity;

void square(int distance) {
	int i;

	barrier;
	for(i=$;i<N;i+=__STARTED_PROCS__) 
		A[i] = B[i] = 0;
	for(i=$;i<distance;i+=__STARTED_PROCS__) 
		D[i] = 0;
	for(i=distance+$;i<N;i+=__STARTED_PROCS__) {
		D[i] = distance;
	}
	barrier;
}

void tree(int arity) {
	int i;

	barrier;
	for(i=$;i<N;i+=__STARTED_PROCS__) 
		A[i] = B[i] = 0;
	if (arity)
	for(i=$;i<N;i+=__STARTED_PROCS__) {
		D[i] = i-i/arity;
	}
	else {
		D[0] = 0;
		for(i=$+1;i<N;i+=__STARTED_PROCS__) {
			D[i] = 1;
		}
	}
	barrier;
}
main(void) {
	int i;
	char c;

	unsigned int localtime, starttime;

	if ($ == 0) {
		printf("\nEnter loop size: ");
		scanf("%d",&N);
		assert(N<=MAXSIZE);
		printf("\nEnter distance dependence: ");
		scanf("%d",&distance);
		assert(distance<=N);
		printf("\nEnter arity dependence: ");
		scanf("%d",&arity);
		printf("\nProcessors = %d N = %d distance = %d arity = %d\n",
					__STARTED_PROCS__,N,distance,arity);
	}
	square(distance);
	if ($ == 0) {

		starttime = getct();
		for(i=0;i<N;i++)
			A[i] = B[i]+A[i-D[i]];
		localtime = getct()-starttime;
		printf("Only 0 working. Sequential Time = %d\n",localtime);
	}

	square(distance);
	slowest = 0;
	barrier;

	starttime = getct();
	for(i=0;i<N;i++)
		A[i] = B[i]+A[i-D[i]]; /*now all the processors are concurrently accesing the shared memory */ 
	localtime = getct()-starttime;
	syncmax(&slowest,localtime);
	pprintf("Sequential Time = %d\n",localtime);
	/* to assure that all the processors have written in slowest */
	barrier; 
	if (!$) pprintf("Slowest Sequential Time = %d\n",slowest);

	square(distance);
	slowest = 0;
	barrier;
	starttime = getct();
	for(i=$; i<N; i+= __STARTED_PROCS__)
		WAIT[i] = (D[i] != 0);
	iteration = 0;
	barrier;
	for(i=mpadd(&iteration,1); i<N; i=mpadd(&iteration,1)) {
		while (WAIT[i-D[i]]) ;
		A[i] = B[i]+A[i-D[i]];
		WAIT[i] = 0;
	}
	localtime = getct()-starttime;
	syncmax(&slowest,localtime);
	pprintf("Parallel Square Time = %d\n",localtime);
	barrier;
	if (!$) {
		pprintf("Slowest Parallel Square Time = %d\n",slowest);
		pprintf("press <return> to continue"); scanf("%c",&c);
	}
	barrier;

	if (!$) pprintf("\n***********************************************\n");
	tree(arity);
	slowest = 0;
	barrier;
	starttime = getct();
	for(i=$; i<N; i+= __STARTED_PROCS__)
		WAIT[i] = (D[i] != 0);
	iteration = 0;
	barrier;
	for(i=mpadd(&iteration,1); i<N; i=mpadd(&iteration,1)) {
		while (WAIT[i-D[i]]) ;
		A[i] = B[i]+A[i-D[i]];
		WAIT[i] = 0;
	}
	localtime = getct()-starttime;
	syncmax(&slowest,localtime);
	pprintf("Parallel Tree Time = %d\n",localtime);
	barrier;
	if (!$) pprintf("Slowest Parallel Tree Time = %d\n",slowest);
	barrier;
}
